/*! ********************************************
 * MultiTimberSynthBase.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_MultiTimberSynthBase_h
#define INCLUDE_MultiTimberSynthBase_h

#ifndef INCLUDE_GrayBox_h
#include "GrayBox.hpp"
#endif

#ifndef INCLUDE_SynthData_h
#include "SynthData.hpp"
#endif

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	VoicePart
************************************************/
class VoicePart
{
public:
	// status
	int32 mode = 0;
	int32 defaultMode = 0;
	// Parameter
	int32 programChange = 0;
	int32 chPressure = 0;
	int32 pitchBend = 8192;
	int32 ccValues[120]={0};
	// Extra
	int32 bendSens = 2;
	int32 extVal0 = 0;
	int32 extVal1 = 0;
	// RPN
	int32 rpn = 0;
	int32 nrpn = 0;
	int32 paramNumber = 0x100;
	
	void reset();
	void resetAllControlle();
	void resetRPNControlle();
// Constructor
	VoicePart(int32 defaultMode = 0);
};
/***********************************************
	Class	MultiTimberSynthBase
************************************************/
class MultiTimberSynthBase : public IGrayBoxModule
{
/***********************************************
	Class	MultiTimberSynthBase :: VoiceUnitBase
************************************************/
protected:
	class VoiceUnitBase
	{
	protected:
	// Tables
		// Volume, Pan
		float* _mainVolTable = nullptr;
		float* _exprVolTable = nullptr;
		float* _veloVolTable = nullptr;
		float* _panTable     = nullptr;
	// ptr
	public:
		void (VoiceUnitBase::*process)(float* const&, float* const&, const int32&) = nullptr;
	// Synthesizer Data
	protected:
		float _masterVolume = 0.0f;
	// VoiceUnit Status
		/*
		- VoiceUnit State -
					 [ FEDC BA98 7654 3210 ] 16Bits Flag.
			value -> [ ---- ---- -KKK dcba ] "-" = Reserved Bit.
			(All Bits "0" = VoiceUnit is DeActive.)
		
			- Operator State -
						 [ 3210 ]
				value -> [ dcba ]
				a = Operator 0
				b = Operator 1
				c = Operator 2
				d = Operator 3
			
			- VoiceUnit Mode - (Exclusive BitFlag)
						 [ FEDC BA98 7654 ]
				value -> [ ---- ---- -KKK ]
					  -> [ ---- ---- -cba ]
			
				a = Key On
				b = Key Off
				c = Sonud Off
		*/
	public:
		int32 state      = 0;
		int32 partNumber = -1;
		int32 noteNumber = 0;
		int32 startFrame = 0;  // Offset Start Sample Frame
		// Volume
	protected:
		int32 _prPan  = 64;  // preset Pan   (0-127)
		float _vPanL  = 0.0f; // total pan L      (val = _panTable[      _prPan + Part.ccPan])
		float _vPanR  = 0.0f; // total pan R      (val = _panTable[256 - _prPan - Part.ccPan])
		float _vPtVol = 0.0f; // CC Volume        (val = _mainVolTable[Part.volume])
		float _vPtExp = 0.0f; // CC Expression    (val = _exprVolTable[Part.expression])
		float _vVel   = 0.0f; // velocity         (val = _veloVolTable[velocity])
		float _vVol   = 0.0f; // etc              (val = _vVel * masterVolume)
		float _vOut   = 0.0f; // output volume    (val = _vPtVol * _vPtExp * _vVol)
		float _vOutL  = 0.0f; // output volume L  (val = _vOut * _vPanL)
		float _vOutR  = 0.0f; // output volume R  (val = _vOut * _vPanR)
	public:
		virtual void killProcess(){ state = 0; partNumber = -1; }
		virtual void keyOff() = 0;
		virtual void soundOff() = 0;
		virtual void keyOn(const int32&, const int32&, const int32&){}
		virtual void setMasterVolume(const float&);
		virtual void updPartBendSens(const int32&) {}
		virtual void updPartPitchBend(const int32&) {}
		virtual void updPartModulation(const int32&) {}
		virtual void updPartVolume(const int32&);
		virtual void updPartExpression(const int32&);
		virtual void updPartPan(const int32&);
	protected:
	// process ---
		virtual void _proc0(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc1(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc2(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc3(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc4(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc5(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc6(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc7(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc8(float* const&, float* const&, const int32&){killProcess();}
		virtual void _proc9(float* const&, float* const&, const int32&){killProcess();}
		virtual void _procA(float* const&, float* const&, const int32&){killProcess();}
		virtual void _procB(float* const&, float* const&, const int32&){killProcess();}
		virtual void _procC(float* const&, float* const&, const int32&){killProcess();}
		virtual void _procD(float* const&, float* const&, const int32&){killProcess();}
		virtual void _procE(float* const&, float* const&, const int32&){killProcess();}
		virtual void _procF(float* const&, float* const&, const int32&){killProcess();}
		// pointer array (process Function)
		void (VoiceUnitBase::*_PROC_PTRS[16])(float* const&, float* const&, const int32&)
			{ &VoiceUnitBase::_proc0, &VoiceUnitBase::_proc1, &VoiceUnitBase::_proc2, &VoiceUnitBase::_proc3,
			  &VoiceUnitBase::_proc4, &VoiceUnitBase::_proc5, &VoiceUnitBase::_proc6, &VoiceUnitBase::_proc7,
			  &VoiceUnitBase::_proc8, &VoiceUnitBase::_proc9, &VoiceUnitBase::_procA, &VoiceUnitBase::_procB,
			  &VoiceUnitBase::_procC, &VoiceUnitBase::_procD, &VoiceUnitBase::_procE, &VoiceUnitBase::_procF };

	};

// Factory -----
protected:
	virtual void _recreateVU(int maxPoly) = 0;
	virtual void _resetKeyParams(){}
// Tables -----
	float _MASTER_VOLUME_TABLE[128]{0.0f};  // for sysEx, NRPN
	virtual void _resetVolumeTables();
// Synthesizer - Data -----
public:
	virtual void factoryDataReset(){}
// Synthesizer - Main -----
	// Synth Status
	int32 midiInPorts  = 0xFFFF;
	int32 midiOutPorts = 0xFFFF;
	bool receiveSysEx = false;
	// Channel Paramters
	VoicePart* VOICE_PARTS = nullptr;
protected:
	float _sampleRate     = 44100.0f;
	float _referencePitch = 440.0f;
	float _masterVolume   = 1.0f;
	// VoiceUnits
	int32 _maxPolyphony = 0;
	VoiceUnitBase** _voiceUnits = nullptr;
	int32 _nton  = 0;
	int32 _ntonM = 0;
	// getter, setter ---
public:
	virtual float getMasterVolume() { return _masterVolume; }
	virtual void setMasterVolume(float val);
	virtual float getSampleRate(){ return _sampleRate; }
	virtual void setSampleRate(float val);
	virtual float getReferencePitch(){ return _referencePitch; }
	virtual void setReferencePitch(float val);
	virtual int32 getPolyphony();
	virtual int32 getMaxPolyphony(){ return _maxPolyphony; };
	virtual int32 setMaxPolyphony(int32 num);
// Constructor
protected:
	MultiTimberSynthBase(int32 maxPoly, float refPitch, float masVol);
// Copy Constructor
	MultiTimberSynthBase(const MultiTimberSynthBase&) = delete;
	MultiTimberSynthBase& operator = (const MultiTimberSynthBase&) = delete;
// Move Constructor
	MultiTimberSynthBase(MultiTimberSynthBase&&) noexcept;
	MultiTimberSynthBase& operator = (MultiTimberSynthBase&&) = delete;
// Destructor
	virtual ~MultiTimberSynthBase();
// Controls ---
public:
	virtual void initialize();
	virtual int32 getDefaultPartMode(int32 part) { return VOICE_PARTS[part].defaultMode; }
	virtual void setDefaultPartMode(int32 part, int32 mode) { VOICE_PARTS[part].defaultMode = mode; }
	virtual int32 getPartMode(int32 part) { return VOICE_PARTS[part].mode; }
	virtual void setPartMode(int32 part, int32 mode);
	virtual void allPartReset(bool modeReset = false);
	virtual void allSoundStop();
	virtual void allSoundOff();
// Interface - IfGrayBoxModule -----
	virtual void updateSampleRate(float) override;
	virtual void updateAudioChannels(int32) override {}
	virtual void updateProcessBlockSize(int32) override {}
	
	virtual void resetModule() override;
	virtual void initiModule() override;
	virtual void connectModule(GrayBox*) override;
	virtual void disconnectModule() override;
	virtual void updateBypassStatus(bool) override;
	
	virtual void startProcess() override;
	virtual void stopProcess() override;
	virtual void suspendProcess() override;
	virtual void resumeProcess() override;
	virtual void process(ProcessData*) override;
// MidiCtrl ---
private:
	MidiEvent _evBuff;
public:
	virtual void noteOff(const int32& ch, const int32& key);
	virtual void noteOn(const int32& ch, const int32& key, const int32& velo);
	virtual void keyAfterTouch(const int32& ch, const int32& key, const int32& val);
	virtual void controleChange(const int32& ch, const int32& cc, const int32& val);
	virtual void programChange(const int32& ch, const int32& num);
	virtual void channelAfterTouch(const int32& ch, const int32& val);
	virtual void pitchBendChange(const int32& ch, const int32& lsb, const int32& msb);
	// val[0 - 16383]
	virtual void pitchBendChangeW(const int32& ch, const int32& val);
	virtual void sysExMessage(const uint8* msg, const int32& dataSize);
// MidiEvent ---
protected:
	virtual void _noteOff(const MidiEvent&);
	virtual void _noteOn(const MidiEvent&);
	virtual void _keyAfterTouch(const MidiEvent&){}
	virtual void _programChange(const MidiEvent&);
	virtual void _channelAfterTouch(const MidiEvent&);
	virtual void _pitchBendChange(const MidiEvent&);
	virtual void _readSysExData(const MidiEvent&){}
// MidiEvent - CC -----
protected:
	// Function Detail : function(ev:MidiEvent):void
	void (MultiTimberSynthBase::*_CC_FUNC[128])(const MidiEvent&);
private:
	void _set_ccFunc();
	void _ccFunc_channelVoiceUB(const MidiEvent&);
	void _ccFunc_channelModeUB(const MidiEvent&){}
	void _ccFunc_01(const MidiEvent&);
	void _ccFunc_06(const MidiEvent&);
	void _ccFunc_07(const MidiEvent&);
	void _ccFunc_0A(const MidiEvent&);
	void _ccFunc_0B(const MidiEvent&);
	void _ccFunc_26(const MidiEvent&);
	void _ccFunc_60(const MidiEvent&);
	void _ccFunc_61(const MidiEvent&);
	void _ccFunc_62(const MidiEvent&);
	void _ccFunc_63(const MidiEvent&);
	void _ccFunc_64(const MidiEvent&);
	void _ccFunc_65(const MidiEvent&);
	void _ccFunc_78(const MidiEvent&);
	void _ccFunc_79(const MidiEvent&);
	void _ccFunc_7B(const MidiEvent&);
// RPN,NRPN -----
	/*
	- RPN,NRPN -
		         [FE DCBA987 6543210]
		value -> [dc bbbbbbb aaaaaaa]
		a = RPN,NRPN (LSB)
		b = RPN,NRPN (MSB)
		c = Flag (Received LSB)
		d = Flag (Received MSB)
			
		paramNumber -> RPN[0-127], NRPN[128-255], nullRPN[256]
		value <= 100h [100h : null RPN] 
		
		_RPN_FUNC
		Function Detail : function(ch:int, param:int, mask:int, inc:int):void
		
	*/
protected:
	void (MultiTimberSynthBase::*_RPN_FUNC[257])(const int32&, const int32&, const int32&, const int32&);
	bool _rpnCtrlFlag  = true;
	bool _nrpnCtrlFlag = false;
public:
	virtual bool getRPNCtrlFlag(){ return _rpnCtrlFlag; }
	virtual void setRPNCtrlFlag(bool flag);
	virtual bool getNRPNCtrlFlag(){ return _nrpnCtrlFlag; }
	virtual void setNRPNCtrlFlag(bool flag);
private:
	void _set_rpnFunc();
	void _rpnFunc_UB(const int32&, const int32&, const int32&, const int32&){}
	void _rpnFunc_Null(const int32&, const int32&, const int32&, const int32&){}
	// bendSens
	void _rpnFunc_00(const int32&, const int32&, const int32&, const int32&);
	// masterTune (FineTune)
	void _rpnFunc_01(const int32&, const int32&, const int32&, const int32&){}
	// masterTune (CoarseTune)
	void _rpnFunc_02(const int32&, const int32&, const int32&, const int32&){}
	// masterVolume
	void _nrpnFunc_00(const int32&, const int32&, const int32&, const int32&);
};

}}}
#endif