/*! ********************************************
 * SynthData.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthData_h
#include "SynthData.hpp"
#endif

#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
// Instance ---
SynthData* SynthData::_instance = nullptr;
SynthData* SynthData::getInstance()
{
	return (_instance != nullptr)? _instance : _instance = new SynthData();
}
void SynthData::disposeInstance()
{
	delete _instance;
	_instance = nullptr;
}
SynthData::~SynthData()
{
// delete ---
	// Volume, Pan
	if (_VOLUME_TABLE_0dB != nullptr)delete[] _VOLUME_TABLE_0dB;
	_VOLUME_TABLE_0dB = nullptr;
	if (_VOLUME_TABLE_6dB != nullptr)delete[] _VOLUME_TABLE_6dB;
	_VOLUME_TABLE_6dB = nullptr;
	if (_PAN_TABLE_SIN != nullptr)delete[] _PAN_TABLE_SIN;
	_PAN_TABLE_SIN = nullptr;
	if (_PAN_TABLE_BAL != nullptr)delete[] _PAN_TABLE_BAL;
	_PAN_TABLE_BAL = nullptr;
	// Log2 Volume Table
	if (_LOG2V_TABLE != nullptr)delete[] _LOG2V_TABLE;
	_LOG2V_TABLE = nullptr;
	if (_LOG2V_TABLE_I != nullptr)delete[] _LOG2V_TABLE_I;
	_LOG2V_TABLE_I = nullptr;
	// Pow2 Table
	if (_POW2_TABLE != nullptr)delete[] _POW2_TABLE;
	_POW2_TABLE = nullptr;
	if (_POW2T_RS_TABLE != nullptr)delete[] _POW2T_RS_TABLE;
	_POW2T_RS_TABLE = nullptr;
	if (_POW2T_KEY_TABLE != nullptr)delete[] _POW2T_KEY_TABLE;
	_POW2T_KEY_TABLE = nullptr;
	// Envelope Generator
	if (_EG_TOTAL_LEVEL_TABLE != nullptr)delete[] _EG_TOTAL_LEVEL_TABLE;
	_EG_TOTAL_LEVEL_TABLE = nullptr;
	if (_EG_LEVEL_TABLE_UP != nullptr)delete[] _EG_LEVEL_TABLE_UP;
	_EG_LEVEL_TABLE_UP = nullptr;
	if (_EG_LEVEL_TABLE_DW != nullptr)delete[] _EG_LEVEL_TABLE_DW;
	_EG_LEVEL_TABLE_DW = nullptr;
	if (_EG_KEY_TABLE != nullptr)delete[] _EG_KEY_TABLE;
	_EG_KEY_TABLE = nullptr;
	if (_EG_POW2T_RS_TABLE != nullptr)delete[] _EG_POW2T_RS_TABLE;
	_EG_POW2T_RS_TABLE = nullptr;
	if (_EG_CNT_GATE_TABLE != nullptr)delete[] _EG_CNT_GATE_TABLE;
	_EG_CNT_GATE_TABLE = nullptr;
	// LFO
	if (_LFO_WAVE_TABLES != nullptr)delete[] _LFO_WAVE_TABLES;
	_LFO_WAVE_TABLES = nullptr;
	// Phase Modulation
	if (_MUL_TABLE != nullptr)delete[] _MUL_TABLE;
	_MUL_TABLE = nullptr;
}
// getter, creator ---
// Volume, Pan ==================================================
float* SynthData::_createPAN_TABLE(int32 type)
{
	float* _table = new float[257];
	switch (type)
	{
		// Panning Law (Balance Pan)
		case 1:
			for (int i = 0; i < 64; ++i)_table[i] = 1.0f;
			for (int i = 64, _v = 0; i < 192; ++i, ++_v)_table[i] = 1.0f - _v / 64.0f;
			for (int i = 192; i < 257; ++i)_table[i] = 0.0f;
			break;
		// Panning Law (Combine Pan)
		case 2:
			for (int i = 0; i < 64; ++i)_table[i] = 1.0f;
			for (int i = 64, _v = 0; i < 192; ++i, ++_v)_table[i] = (128-_v) / 128.0f;
			for (int i = 192; i < 257; ++i)_table[i] = 0.0f;
			break;
		// Panning Law (Sine Curve)
		default:
			const float _PIS2f = 3.141592653589793f / 2.0f;
			for (int i = 0; i < 64; ++i)_table[i] = 1.0f;
			for (int i = 64, _v = 0; i < 192; ++i, ++_v)_table[i] = std::sinf(((128 - _v) / 128.0f) * _PIS2f);
			for (int i = 192; i < 257; ++i)_table[i] = 0.0f;
			break;
	}
	return _table;
}
float* SynthData::createVOLUME_TABLE(int32 size, float c, bool reverse)
{
	int i = 0, j = 1, _cnt = 1;
	float _sizeF = float(size);
	float* _table = new float[size];
	if (reverse) { j = size; _cnt = -1; }
	for (i = 0; i < size; ++i, j += _cnt)
	{
		float _nv = std::powf(j / _sizeF, 2.0f) * c;
		_table[i] = _nv;
	}
	return _table;
}
float* SynthData::getVOLUME_TABLE_0dB(){ return (_VOLUME_TABLE_0dB != nullptr)? _VOLUME_TABLE_0dB : _VOLUME_TABLE_0dB = createVOLUME_TABLE(128, 1.0f); } 
float* SynthData::getVOLUME_TABLE_6dB(){ return (_VOLUME_TABLE_6dB != nullptr)? _VOLUME_TABLE_6dB : _VOLUME_TABLE_6dB = createVOLUME_TABLE(128, 0.5011872336272722f); } 
float* SynthData::getPAN_TABLE_SIN(){ return (_PAN_TABLE_SIN != nullptr)? _PAN_TABLE_SIN : _PAN_TABLE_SIN = _createPAN_TABLE(0); } 
float* SynthData::getPAN_TABLE_BAL(){ return (_PAN_TABLE_BAL != nullptr)? _PAN_TABLE_BAL : _PAN_TABLE_BAL = _createPAN_TABLE(1); } 
// Log2 Volume Table ==================================================
int32 SynthData::toLog2vtIndexValue(float vol)
{
	if      (vol >  _LOG2VT_BOADERf) return  int(std::log2f( vol) * _LOG2VT_DYN_RESf + 0.5f) << 1;
	else if (vol < -_LOG2VT_BOADERf) return (int(std::log2f(-vol) * _LOG2VT_DYN_RESf + 0.5f) << 1) + 1;  // +1 = Nega
	else return LOG2VT_SIZE;
}
int32 SynthData::toLog2vtIndexValue(double vol)
{
	if      (vol >  _LOG2VT_BOADERd) return  int(std::log2( vol) * _LOG2VT_DYN_RESd + 0.5) << 1;
	else if (vol < -_LOG2VT_BOADERd) return (int(std::log2(-vol) * _LOG2VT_DYN_RESd + 0.5) << 1) + 1;  // +1 = Nega
	else return LOG2VT_SIZE;
}
float* SynthData::getLOG2V_TABLE()
{
	if (_LOG2V_TABLE != nullptr) return _LOG2V_TABLE;
	const float _dynB = float(LOG2VT_DYN_BITS),
		_maxV = float(LOG2VT_DYN), _res = float(LOG2VT_DYN_RES);
	const int _tableSize = (LOG2VT_SIZE + 2) * 4;
	float* _table = new float[_tableSize];
	for (int i = 0, j = 0; i < LOG2VT_SIZE; ++j)
	{
		float _nv = std::powf(2.0f, _dynB - j / _res) / _maxV;
		_table[i++] =  _nv;
		_table[i++] = -_nv;
	}
	for (int i = LOG2VT_SIZE; i < _tableSize; ++i)_table[i] = 0.0f;
	return _LOG2V_TABLE = _table;
}
int32* SynthData::getLOG2V_TABLE_I()
{
	if (_LOG2V_TABLE_I != nullptr) return _LOG2V_TABLE_I;
	const double _res = double(LOG2VT_DYN_RES), _dynB = double(LOG2VT_DYN_BITS);
	const int _tableSize = (LOG2VT_SIZE + 2) * 4;
	int32* _table = new int32[_tableSize];
	for (int i = 0, j = 0; i < LOG2VT_SIZE; ++j)
	{
		int _iv = int(std::pow(2.0, _dynB - j / _res));
		_table[i++] =  _iv;
		_table[i++] = -_iv;
	}
	for (int i = LOG2VT_SIZE; i < _tableSize; ++i)_table[i] = 0;
	return _LOG2V_TABLE_I = _table;
}

// Pow2 Table ==================================================
int32 SynthData::toPow2tIndexValue(float val)
{
	if (val <= 1.0f) return 0;
	else return int(std::log2f(val) * _POW2T_OCTV_RESf + 0.5f);
}
int32 SynthData::toPow2tIndexValue(double val)
{
	if (val <= 1.0) return 0;
	else return int(std::log2(val) * _POW2T_OCTV_RESd + 0.5);
}
int32* SynthData::getPOW2_TABLE()
{
	if (_POW2_TABLE != nullptr) return _POW2_TABLE;
	double _octvRes = double(POW2T_OCTV_RES);
	int32* _table = new int32[POW2T_OCTV_RES];
	for (int i = 0; i < POW2T_OCTV_RES; ++i)_table[i] = int(std::pow(2.0, 30.0 + (i / _octvRes)) + 0.5);
	return _POW2_TABLE = _table;
}
int32* SynthData::getPOW2T_RS_TABLE()
{
	if (_POW2T_RS_TABLE != nullptr) return _POW2T_RS_TABLE;
	int32* _table = new int32[128];
	for (int i =  0; i <  30; ++i)_table[i] = 30 - i;
	for (int i = 30; i < 128; ++i)_table[i] = 0;  // Safe Field
	return _POW2T_RS_TABLE = _table;
}
int32* SynthData::getPOW2T_KEY_TABLE()
{
	if (_POW2T_KEY_TABLE != nullptr) return _POW2T_KEY_TABLE;
	double _octvRes = double(POW2T_OCTV_RES);
	int32* _table = new int32[POW2T_OCTV_RES];
	for (int i = 0; i < 128; ++i)_table[i] = int(i / 12.0 * _octvRes + 0.5);
	return _POW2T_KEY_TABLE = _table;
}
// Envelope Generator ==================================================
int32* SynthData::getEG_TOTAL_LEVEL_TABLE()
{
	if (_EG_TOTAL_LEVEL_TABLE != nullptr) return _EG_TOTAL_LEVEL_TABLE;
	// Ref - YM2151
		// D6 D5 D4 D3 D2 D1  D0
		// 48 24 12 6  3  1.5 0.75 [dB]
	int _step = LOG2VT_SIZE >> 7;
	int32* _table = new int32[128*2];
	for (int i =   0; i < 128; ++i)_table[i] = i * _step;  // 64(0.75dB) = 8192/128
	for (int i = 128; i < 256; ++i)_table[i] = LOG2VT_SIZE;  // VeloCtrl
	return _EG_TOTAL_LEVEL_TABLE = _table;
}
int32* SynthData::getEG_LEVEL_TABLE_DW()
{
	if (_EG_LEVEL_TABLE_DW != nullptr) return _EG_LEVEL_TABLE_DW;
	int32* _table = new int32[EG_MAX_LEVEL];
	for (int i = 0; i < EG_MAX_LEVEL; ++i)_table[i] = i << 1;
	return _EG_LEVEL_TABLE_DW = _table;
}
int32* SynthData::getEG_LEVEL_TABLE_UP()
{
	if (_EG_LEVEL_TABLE_UP != nullptr) return _EG_LEVEL_TABLE_UP;
	double _size = EG_MAX_LEVEL;
	int32* _table = new int32[EG_MAX_LEVEL];
	for (int i = 0; i < EG_MAX_LEVEL; ++i)_table[i] = toLog2vtIndexValue(i / _size);
	return _EG_LEVEL_TABLE_UP = _table;
}
int32* SynthData::getEG_KEY_TABLE()
{
	if (_EG_KEY_TABLE != nullptr) return _EG_KEY_TABLE;
	double _octvRes = POW2T_OCTV_RES;
	int32* _table = new int32[256];
	for (int i = 0; i < 128; ++i)_table[i] = 0;
	for (int i = 128, _b = 1; i < 256; ++i, ++_b)_table[i] = int(_b / 48.0 * _octvRes);
	return _EG_KEY_TABLE = _table;
}
int32 * SynthData::getEG_POW2T_RS_TABLE()
{
	if (_EG_POW2T_RS_TABLE != nullptr) return _EG_POW2T_RS_TABLE;
	int32* _table = new int32[128];
	for (int i =  0; i <= EG_MAX_CNT_BITS; ++i)_table[i] = 30 - i;
	for (int i = EG_MAX_CNT_BITS; i < 128; ++i)_table[i] = 31 - EG_MAX_CNT_BITS; // Safe Field (value >= 1)
	return _EG_POW2T_RS_TABLE = _table;
}
int32 * SynthData::getEG_CNT_GATE_TABLE()
{
	if (_EG_CNT_GATE_TABLE != nullptr) return _EG_CNT_GATE_TABLE;
		// D6 D5 D4 D3 D2 D1  D0
		// 48 24 12 6  3  1.5 0.75 [dB]
	int _btm = EG_ENV_DYN_BITS * 8, 
		_step = (EG_MAX_LEVEL << EG_CNT_FP) / 128;
	int32* _table = new int32[129];
	for (int i =    0; i < _btm; i++)_table[i] = i * _step;
	for (int i = _btm; i <  129; i++)_table[i] = EG_CNT_ENV_BTM;
	return _EG_CNT_GATE_TABLE = _table;
}
// LFO ==================================================
int32 SynthData::lfoTimerCoef(float sampleRate, int32 steps)
{
	float _tableSize = LFO_WAVE_TABLE_SIZE / 2.0f,
		_lfoMaxFrameRate = sampleRate / LFO_MAX_FREQf / _tableSize;
	if (_lfoMaxFrameRate < 1.0f)_lfoMaxFrameRate = 1.0f;  // Max 1frame per 1Phase
	return int32(LFO_TIMER_INIT_VALUE / float(steps) / _lfoMaxFrameRate + 0.5f);
}
int32 SynthData::lfoTimerCoef(double sampleRate, int32 steps)
{
	double _tableSize = LFO_WAVE_TABLE_SIZE / 2.0,
		_lfoMaxFrameRate = sampleRate / LFO_MAX_FREQ / _tableSize;
	if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
	return int32(LFO_TIMER_INIT_VALUE / double(steps) / _lfoMaxFrameRate + 0.5);
}
int32* SynthData::getLFO_WAVE_TABLES()
{
	if (_LFO_WAVE_TABLES != nullptr) return _LFO_WAVE_TABLES;
	const int _SIZ = LFO_WAVE_TABLE_SIZE, _HLF = _SIZ >> 1,
	          _QTR = _SIZ >> 2, _TRI = _HLF + _QTR;
	const double _SIZ_N = double(_SIZ), _HLF_N = _SIZ_N / 2.0, _QTR_N = _SIZ_N / 4.0;
	const double _LOG2VT_SIZEd = double(LOG2VT_SIZE), _POW2T_OCTV_RESd = double(POW2T_OCTV_RES);
	int32*_tables = new int32[_SIZ * 4];
	// Tri ---
	int *_table = &_tables[0];
		// PM
		for (int i = 0; i < _QTR; i += 2)
		{
			int _iv = int(i / _QTR_N * _POW2T_OCTV_RESd);
			_table[       i] = _iv;
			_table[_QTR + i] = POW2T_OCTV_RES - _iv;
			_table[_HLF + i] = -_iv;
			_table[_TRI + i] = _iv - POW2T_OCTV_RES;
		}
		// AM
		for (int i = 0; i < _HLF; i+=2)	
		{
			int _iv = int(i / _HLF_N * _LOG2VT_SIZEd);
			_table[       i + 1] = _iv;
			_table[_HLF + i + 1] = LOG2VT_SIZE - _iv;
		}
	// Rect ---
	_table = &_tables[_SIZ];
		for (int i = 0; i < _HLF; i+=2)	
		{
			// PM	
			_table[       i] =  POW2T_OCTV_RES;
			_table[_HLF + i] = -POW2T_OCTV_RES;
			// AM
			_table[       i + 1] = 0;
			_table[_HLF + i + 1] = LOG2VT_SIZE;
		}
	// Saw ---
	_table = &_tables[_SIZ * 2];
		for (int i = 0; i < _SIZ; i+=2)	
		{
			// PM	
			_table[i    ] = POW2T_OCTV_RES - int(i / _HLF_N * _POW2T_OCTV_RESd);
			// AM
			_table[i + 1] = int(i / _SIZ_N * _LOG2VT_SIZEd);
		}
	// Noise ---
	_table = &_tables[_SIZ * 3];
	std::srand(1635);
	for (int i = 0; i < _SIZ; i+=2)	
	{
		double _nv = (std::rand() & 32767) / 32767.0;
		// AM
		_table[i + 1] = int(_nv * _LOG2VT_SIZEd);
		// PM
		_nv = _nv * 2.0 - 1.0;
		_table[i    ] = int(_nv * _POW2T_OCTV_RESd);
	}
	return _LFO_WAVE_TABLES = _tables;
}
// Phase Modulation ==================================================
int32* SynthData::getMUL_TABLE()
{
	if (_MUL_TABLE != nullptr) return _MUL_TABLE;
	const int _octvRes = POW2T_OCTV_RES;
	int32* _table = new int32[16];
	for (int i = 0; i < 16; ++i)_table[i] = (i-1) * _octvRes;
	return _MUL_TABLE = _table;
}
}}}