/*! ********************************************
 * SynthesizerFM2.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerFM2_h
#include "SynthesizerFM2.hpp"
#endif

#include <utility>
#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	SynthesizerFM2
************************************************/
void SynthesizerFM2::_recreateVU(int32 maxPoly)
{
	// Recreate Array
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = new VoiceUnitBase*[maxPoly];
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = new VoiceUnitFM2[maxPoly];
	// Reset VoiceUnit
	for (int i = 0; i < maxPoly; ++i)
	{
		VoiceUnitFM2& _vu = _voiceUnitExt[i];
		_voiceUnits[i] = &_voiceUnitExt[i];
		_vu.setup(_synthData);
		_vu.killProcess();
		_vu.VOICE_PARTS = VOICE_PARTS;
		_vu.WAVE_TABLES = WAVE_TABLES;
		_vu.SAMPLE_DATAS = SAMPLE_DATAS;
		_vu.TONE_PRESETS = TONE_PRESETS;
		_vu.SAMPLER_PRESETS = SAMPLER_PRESETS;
		_vu.setMasterVolume(_masterVolume);
	}
	// set Property
	_maxPolyphony = maxPoly;
}
void SynthesizerFM2::_resetKeyParams()
{
	const double _sampleRateD = double(_sampleRate);
	// LFO
	double _lfoWaveSize = SynthData::LFO_WAVE_TABLE_SIZE / 2.0;
		// Max Frame Rate - LFO Phase Step
	double _lfoMaxFrameRate = _sampleRateD / SynthData::LFO_MAX_FREQ / _lfoWaveSize;
	if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
	int _lfoTimerCoef = int(SynthData::LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		
	// EG
	double _rate0Frames = SynthData::EG_REFERENCE_TIME * SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egPow2tIdx = SynthData::toPow2tIndexValue(SynthData::EG_MAX_CNT / _rate0Frames);
	double _sndOffFrames = SynthData::EG_SOUND_OFF_TIME / SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egSndOff = int(SynthData::EG_MAX_CNT / _sndOffFrames + 0.5);
	// PG
	int _smpl = SynthData::toPow2tIndexValue(_sampleRateD);
	int _pgSamplePow2tIdx = SynthData::toPow2tIndexValue(double(1 << PG_PHASE_FP)) - _smpl;
	// PG WaveTable
	int _pgKeyPow2tIdx = SynthData::toPow2tIndexValue(double(WAVE_TABLE_SIZE << PG_PHASE_FP)) - _smpl;
	_pgKeyPow2tIdx += SynthData::toPow2tIndexValue(_referencePitch * std::pow(2.0, -69.0 / 12.0)); // note 0 Pitch
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitFM2& _vu = _voiceUnitExt[i];
		_vu.killProcess();
		_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
		_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
		_vu.egPow2tIdx = _egPow2tIdx;
		_vu.egPhaseStepSndOff = _egSndOff;
		_vu.lfoTimerCoef = _lfoTimerCoef;
	}
}
void SynthesizerFM2::getWaveForm(int32 id, float* buffer, int32 bufferSize)
{
	if (bufferSize > WAVE_TABLE_SIZE)bufferSize = WAVE_TABLE_SIZE;
	// ---
	const float* _log2vt = _synthData->getLOG2V_TABLE();
	const int32* _table = &WAVE_TABLES[id << WAVE_TABLE_SIZE_BITS];
	for (int i = 0; i < bufferSize; ++i)buffer[i] = _log2vt[_table[i]];
}
void SynthesizerFM2::setWaveForm(int32 id, const float* buffer, int32 bufferSize)
{
	if (bufferSize > WAVE_TABLE_SIZE)bufferSize = WAVE_TABLE_SIZE;
	const float _RES = float(-SynthData::LOG2VT_DYN_RES),
		_BOADER = 1.0f / (1 << SynthData::LOG2VT_DYN_BITS);
	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	int32* _table = &WAVE_TABLES[id << WAVE_TABLE_SIZE_BITS];
	for (int i = 0; i < bufferSize; ++i)
	{
		float v = buffer[i];
		if      (v >  _BOADER) _table[i] =  int(std::log2f( v) * _RES + 0.5f) << 1;
		else if (v < -_BOADER) _table[i] = (int(std::log2f(-v) * _RES + 0.5f) << 1) + 1;
		else _table[i] = _BOTTOM;
	}
}
void SynthesizerFM2::setSampleData(int32 id, SampleData&& sampleData)
{
	allSoundStop();
	SAMPLE_DATAS[id] = std::move(sampleData);
}
void SynthesizerFM2::setSampleData(int32 id, const SampleData& sampleData, bool deepCopy)
{
	allSoundStop();
	SAMPLE_DATAS[id].copy(sampleData, deepCopy);
}
void SynthesizerFM2::clearSampleData(int32 id)
{
	allSoundStop();
	SAMPLE_DATAS[id].clear();
}
void SynthesizerFM2::factoryDataReset()
{
	if (_voiceUnitExt != nullptr) allSoundStop();
	// waveData
	const double _RES = double(-SynthData::LOG2VT_DYN_RES),
		_BOADER = 1.0 / (1 << SynthData::LOG2VT_DYN_BITS);
	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	const double _PID2 = 3.141592653589793 * 2.0;
	const int _SIZE = WAVE_TABLE_SIZE, _HALF = _SIZE >> 1, _QTER = _SIZE >> 2;
	const double _SIZEd = double(_SIZE), _HALFd = double(_HALF), _QTERd = double(_QTER);
	double _sl[OVER_TONES]{0.0};
	double _sw[OVER_TONES]{0.0};
	
	if (WAVE_TABLES == nullptr)WAVE_TABLES = new int32[8 << WAVE_TABLE_SIZE_BITS];
	int32* _wave0 = WAVE_TABLES;
	int32* _wave1 = &WAVE_TABLES[1 << WAVE_TABLE_SIZE_BITS];
	int32* _wave2 = &WAVE_TABLES[2 << WAVE_TABLE_SIZE_BITS];
	int32* _wave3 = &WAVE_TABLES[3 << WAVE_TABLE_SIZE_BITS];
	int32* _wave4 = &WAVE_TABLES[4 << WAVE_TABLE_SIZE_BITS];
	int32* _wave5 = &WAVE_TABLES[5 << WAVE_TABLE_SIZE_BITS];
	int32* _wave6 = &WAVE_TABLES[6 << WAVE_TABLE_SIZE_BITS];
	int32* _wave7 = &WAVE_TABLES[7 << WAVE_TABLE_SIZE_BITS];
	for (int i = 0, _iv = 0; i < _HALF; ++i)
	{
	// sine [ nu ]
		double _nv = std::sin((i / _SIZEd) * _PID2);
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave0[i] = _iv;
		_wave0[_HALF + i] = _iv + 1;
	// sine [ n- ]
		_wave1[i] = _iv;
		_wave1[_HALF + i] = _BOTTOM;
	// sine [ nn ]
		_wave2[i] = _iv; 
		_wave2[_HALF + i] = _iv;
		if	((i & 1) == 0){ 
	// sine [nu-- ]
			_wave3[i >> 1] = _iv; 
			_wave3[_QTER + (i >> 1)] = _iv + 1;
	// sine [ nu-- ]
			_wave4[i >> 1] = _iv; 
			_wave4[_QTER + (i >> 1)] = _iv; }
		_wave3[_HALF + i] = _BOTTOM;
		_wave4[_HALF + i] = _BOTTOM;
	}
// triangle ----
	for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sw[i] = _SIZEd / j;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sl[i] = (i & 1)? -1.0 / (j * j) : 1.0 / (j * j);
	double _den = 0.0;
	for (int j = 0; j < OVER_TONES; ++j)_den += std::sin(_QTERd / _sw[j] * _PID2) * _sl[j];
	_den = 1.0 / _den;
	for (int i = 0, _iv = 0; i < _SIZE; ++i)
	{
		double _nv = 0.0;
		for (int j = 0; j < OVER_TONES; ++j)_nv += std::sin((i / _sw[j]) * _PID2) * _sl[j];
		_nv *= _den;
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave5[i] = _iv;
	}
// square ----
	//for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sw[i] = _SIZEd / j;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sl[i] = 1.0 / j;
	const double _cSQR = _QTERd / OVER_TONES;
	_den = 0.0;
	for (int j = 0; j < OVER_TONES; ++j)_den += std::sin(_cSQR / _sw[j] * _PID2) * _sl[j];
	_den = 1.0 / _den;
	for (int i = 0, _iv = 0; i < _SIZE; ++i)
	{
		double _nv = 0.0;
		for (int j = 0; j < OVER_TONES; ++j)_nv += std::sin((i / _sw[j]) * _PID2) * _sl[j];
		_nv *= _den;
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave6[i] = _iv;
	}
// saw ----
	for (int i = 0, j = 1; i < OVER_TONES; ++i, ++j)_sw[i] = _SIZE / j;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, ++j)_sl[i] = 1.0 / j;
	double _c = _HALFd;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, ++j) _c /= 1.0 + 1.0 / j;
	_c = std::round(_c + 0.5);
	_den = 0.0;
	for (int j = 0; j < OVER_TONES; ++j)_den += std::sin(_c / _sw[j] * _PID2) * _sl[j];
	_den = 1.0 / _den;
	for (int i = 0, _iv = 0; i < _SIZE; ++i)
	{
		double _nv = 0.0;
		for (int j = 0; j < OVER_TONES; ++j)_nv += std::sin((i / _sw[j]) * _PID2) * _sl[j];
		_nv *= _den;
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave7[i] = _iv;
	}
	// Preset
	if (TONE_PRESETS == nullptr)TONE_PRESETS = new TonePresetFM[256];
	for (int i = 0; i < 256; ++i)TONE_PRESETS[i].reset();
// Sampler
	if (SAMPLE_DATAS == nullptr)SAMPLE_DATAS = new SampleData[128];
	if (SAMPLER_PRESETS == nullptr)SAMPLER_PRESETS = new SamplerPreset[128];
	for (int i = 0; i < 128; ++i)
	{
		SAMPLE_DATAS[i].clear();
		SAMPLER_PRESETS[i].reset();
	}
}
// Constructor
SynthesizerFM2::SynthesizerFM2(int32 maxPoly, float refPitch, float masVol, bool privateTableData) : MultiTimberSynthBase::MultiTimberSynthBase(maxPoly, refPitch, masVol)
{
	// create
	_synthData = (privateTableData) ? _prvSynthData = new SynthData() : SynthData::getInstance();
	factoryDataReset();
	_recreateVU(_maxPolyphony);
	_resetKeyParams();
}
SynthesizerFM2::SynthesizerFM2(SynthesizerFM2 && x) noexcept : MultiTimberSynthBase::MultiTimberSynthBase(std::move(x))
{
	// Synthesizer - Data -----
	_prvSynthData = x._prvSynthData; x._prvSynthData = nullptr;
	_synthData = x._synthData; x._synthData = nullptr;
	WAVE_TABLES = x.WAVE_TABLES; x.WAVE_TABLES = nullptr;
	SAMPLE_DATAS = x.SAMPLE_DATAS; x.SAMPLE_DATAS = nullptr;
	TONE_PRESETS = x.TONE_PRESETS; x.TONE_PRESETS = nullptr;
	SAMPLER_PRESETS = x.SAMPLER_PRESETS; x.SAMPLER_PRESETS = nullptr;
	// Synthesizer - main -----
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;
}
SynthesizerFM2 & SynthesizerFM2::operator=(SynthesizerFM2 && x) noexcept
{
	if (this == &x)return *this;
	// MultiTimberSynthBase::operator=(std::move(x));
// MultiTimberSynthBase ===
	// Synth Status
	midiInPorts = x.midiInPorts;
	midiOutPorts = x.midiOutPorts;
	receiveSysEx = x.receiveSysEx;
	// Channel Paramters
	if (VOICE_PARTS != nullptr)delete[] VOICE_PARTS;
	VOICE_PARTS = x.VOICE_PARTS; x.VOICE_PARTS = nullptr;
	_sampleRate = x._sampleRate;
	_referencePitch = x._referencePitch;
	_masterVolume = x._masterVolume;
	// VoiceUnits
	_maxPolyphony = x._maxPolyphony;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = x._voiceUnits; x._voiceUnits = nullptr;
	_nton = x._nton;
	_ntonM = x._ntonM;
	// RPN,NRPN -----
	_rpnCtrlFlag = x._rpnCtrlFlag;
	_nrpnCtrlFlag = x._nrpnCtrlFlag;
// SynthesizerFM2 ===
	// Synthesizer - Data -----
	_synthData = x._synthData;
	if (_prvSynthData != nullptr)delete _prvSynthData;
	_prvSynthData = x._prvSynthData; x._prvSynthData = nullptr;
	if (WAVE_TABLES != nullptr)delete[] WAVE_TABLES;
	WAVE_TABLES = x.WAVE_TABLES; x.WAVE_TABLES = nullptr;
	if (SAMPLE_DATAS != nullptr)delete[] SAMPLE_DATAS;
	SAMPLE_DATAS = x.SAMPLE_DATAS; x.SAMPLE_DATAS = nullptr;
	if (TONE_PRESETS != nullptr)delete[] TONE_PRESETS;
	TONE_PRESETS = x.TONE_PRESETS; x.TONE_PRESETS = nullptr;
	if (SAMPLER_PRESETS != nullptr)delete[] SAMPLER_PRESETS;
	SAMPLER_PRESETS = x.SAMPLER_PRESETS; x.SAMPLER_PRESETS = nullptr;
	// Synthesizer - main -----
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;

	return *this;
}
SynthesizerFM2::~SynthesizerFM2()
{
	if (VOICE_PARTS != nullptr)delete[] VOICE_PARTS;
	VOICE_PARTS = nullptr;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = nullptr;
	if (WAVE_TABLES != nullptr)delete[] WAVE_TABLES;
	WAVE_TABLES = nullptr;
	if (SAMPLE_DATAS != nullptr)delete[] SAMPLE_DATAS;
	SAMPLE_DATAS = nullptr;
	if (TONE_PRESETS != nullptr)delete[] TONE_PRESETS;
	TONE_PRESETS = nullptr;
	if (SAMPLER_PRESETS != nullptr)delete[] SAMPLER_PRESETS;
	SAMPLER_PRESETS = nullptr;
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = nullptr;
	if (_prvSynthData != nullptr)delete _prvSynthData;
	_prvSynthData = nullptr;
}
// MidiEvent ---
void SynthesizerFM2::_noteOn(const MidiEvent& ev)
{
	VoicePart& _pt = VOICE_PARTS[ev.channel];
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitFM2& _vu = _voiceUnitExt[i];
		if (_vu.state == 0)
		{
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	const int _len = _nton + _maxPolyphony;
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitFM2& _vu = _voiceUnitExt[i & _ntonM];
		// (KeyOff || SoundOff) -->
		if (_vu.state & 0x60)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitFM2& _vu = _voiceUnitExt[i & _ntonM];
		// KeyOn -->
		if (_vu.state & 0x10)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
}
void SynthesizerFM2::_readSysExData(const MidiEvent& ev)
{
	if (ev.dataSize < 8)return;
	const uint8* dat = ev.data;
	const uint8* dataEnd = dat + ev.dataSize - 1;  // minus F7
	if (*(dat++) != 0x7Du)return;  // ManuFactureID
	if (*(dat++) != 0x10u)return;  // DeviceID
	if (*(dat++) != 0x21u)return;  // ModelID

	uint8 _padr = 0u;
	switch (*(dat++))	// Address 1
	{
	// Common
		case 0x00u:
			if (*(dat++) != 0x00u)return; // Address 2
			switch (*(dat++))	// Address 3
			{
				case 0x00u : setMasterVolume(_MASTER_VOLUME_TABLE[*(dat++)]); if (dat >= dataEnd) break;
				case 0x01u : setReferencePitch(440.0f + (*dat - 64.0f) * 0.5f); break;
			}
			break;
	// Inst (Drum map)
		case 0x05u: _padr = 128u;
	// Inst (Program map)
		case 0x04u:{
			TonePresetFM& _preset = TONE_PRESETS[_padr | *(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _preset.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _preset.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _preset.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _preset.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _preset.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _preset.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _preset.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _preset.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _preset.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _preset.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _preset.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _preset.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _preset.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _preset.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _preset.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _preset.toneName[15] = *(dat++);  if (dat >= dataEnd) break;
				
				case 0x10u : ++dat; if (dat >= dataEnd) break;
				case 0x11u : ++dat; if (dat >= dataEnd) break;
				case 0x12u : ++dat; if (dat >= dataEnd) break;
				case 0x13u : ++dat; if (dat >= dataEnd) break;
				case 0x14u : ++dat; if (dat >= dataEnd) break;
				case 0x15u : ++dat; if (dat >= dataEnd) break;
				case 0x16u : ++dat; if (dat >= dataEnd) break;
				case 0x17u : ++dat; if (dat >= dataEnd) break;
				case 0x18u : ++dat; if (dat >= dataEnd) break;
				case 0x19u : _preset.key     = *(dat++);     if (dat >= dataEnd) break;
				case 0x1Au : _preset.pan     = *(dat++);     if (dat >= dataEnd) break;
				case 0x1Bu : _preset.lfoWave = *(dat++) & 3; if (dat >= dataEnd) break;
				case 0x1Cu : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x1Du : _preset.lfoFreq = _preset.lfoFreq & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x1Eu : _preset.fl      = *(dat++);     if (dat >= dataEnd) break;
				case 0x1Fu : _preset.strc    = *(dat++) & 7; if (dat >= dataEnd) break;
				
				case 0x20u : ++dat; if (dat >= dataEnd) break;
				case 0x21u : ++dat; if (dat >= dataEnd) break;
				case 0x22u : ++dat; if (dat >= dataEnd) break;
				case 0x23u : _preset.tlVelCtrl_0 = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x24u : _preset.mul_0       = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x25u : _preset.tune_0      = _preset.tune_0 & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x26u : _preset.tune_0      = _preset.tune_0 & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x27u : _preset.lms_0       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x28u : _preset.wave_0      = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x29u : _preset.tl_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Au : _preset.ar_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Bu : _preset.dr_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Cu : _preset.sl_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Du : _preset.rr_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Eu : _preset.kf_0        = *(dat++) & 3;  if (dat >= dataEnd) break;
				case 0x2Fu : _preset.kfStart_0   = *(dat++);      if (dat >= dataEnd) break;
				
				case 0x30u : ++dat;							 if (dat >= dataEnd) break;
				case 0x31u : ++dat;							 if (dat >= dataEnd) break;
				case 0x32u : ++dat;							 if (dat >= dataEnd) break;
				case 0x33u : _preset.tlVelCtrl_1 = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x34u : _preset.mul_1       = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x35u : _preset.tune_1      = _preset.tune_1 & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x36u : _preset.tune_1      = _preset.tune_1 & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x37u : _preset.lms_1       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x38u : _preset.wave_1      = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x39u : _preset.tl_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Au : _preset.ar_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Bu : _preset.dr_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Cu : _preset.sl_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Du : _preset.rr_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Eu : _preset.kf_1        = *(dat++) & 3;  if (dat >= dataEnd) break;
				case 0x3Fu : _preset.kfStart_1   = *(dat++);      break;
			}
			} break;
	// Percussion
		case 0x06u: {
			SamplerPreset& _smplPrst = SAMPLER_PRESETS[*(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _smplPrst.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _smplPrst.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _smplPrst.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _smplPrst.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _smplPrst.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _smplPrst.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _smplPrst.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _smplPrst.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _smplPrst.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _smplPrst.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _smplPrst.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _smplPrst.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _smplPrst.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _smplPrst.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _smplPrst.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _smplPrst.toneName[15] = *(dat++);  if (dat >= dataEnd) break;

				case 0x10u : _smplPrst.smplNum = *(dat++);       if (dat >= dataEnd) break;
				case 0x11u : _smplPrst.pan     = *(dat++);       if (dat >= dataEnd) break;
				case 0x12u : _smplPrst.octv    = *(dat++) & 15;  if (dat >= dataEnd) break;
				case 0x13u : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x14u : _smplPrst.tune    = _smplPrst.tune & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x15u : _smplPrst.tl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x16u : _smplPrst.ar      = *(dat++);       if (dat >= dataEnd) break;
				case 0x17u : _smplPrst.dr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x18u : _smplPrst.sl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x19u : _smplPrst.rr      = *dat;           break;
			}
			} break;
	}
}
/***********************************************
	Class	SynthesizerFM2 :: VoiceUnitFM2
************************************************/
void SynthesizerFM2::VoiceUnitFM2::keyOff()
{
	// KeyOn -->
	if (state & 0x0010)
	{
	// OP0 is Active -->
		if (state & 1)
		{
			if (_egState_0 == 1)
			{
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
				else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			_egGate_0 = _egCntEnvBtm_0;
			_egCountStep_0 = _egCountStepRR_0;
			_egState_0 = 8;
			if (_egCount_0 > _egCntEnvBtm_0)
			{ 
				if (state & 2){ state &= 0xf2; _egState_0 = 0; }
				else { killProcess(); return; }
			}
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	// OP1 is Active -->
		if (state & 2)
		{
			if (_egState_1 == 1)
			{
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
				else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			_egGate_1 = _egCntEnvBtm_1;
			_egCountStep_1 = _egCountStepRR_1;
			_egState_1 = 8;
			if (_egCount_1 > _egCntEnvBtm_1)
			{ 
				if (state & 1){ state &= 0xf1; _egState_1 = 0; }
				else { killProcess(); return; }
			}
		}
		state = 0x0020 | state & 3;
	}
}
void SynthesizerFM2::VoiceUnitFM2::soundOff()
{
	// KeyOn || KeyOff --->
	if (state & 0x0030)
	{
	// OP0 Active --->
		if (state & 1)
		{
			if (_egState_0 == 1)
			{
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
				else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			if (_egCount_0 > _egCntEnvBtm_0)
			{ 
				if (state & 2){ state &= 0xf2; }
				else { killProcess(); return; }
			}
			_egGate_0 = _egCntEnvBtm_0;
			_egCountStep_0 = egPhaseStepSndOff;
			_egState_0 = 0;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	// OP1 Active --->
		if (state & 2)
		{
			if (_egState_1 == 1)
			{
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
				else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			if (_egCount_1 > _egCntEnvBtm_1)
			{ 
				if (state & 1){ state &= 0xf1; }
				else { killProcess(); return; }
			}
			_egGate_1 = _egCntEnvBtm_1;
			_egCountStep_1 = egPhaseStepSndOff;
			_egState_1 = 0;
		}
		state = 0x0040 | state & 3;
		partNumber = -1; // SoundOff
	}
}
int32 SynthesizerFM2::VoiceUnitFM2::_shiftEGState_0()
{
	switch (_egState_0)
	{
		// ATTACK ->
		case 1:
			_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
			_egCount_0 = 0;
			// -> DECAY
			if (_egSL_0 > 0)
			{
				_egGate_0 = _egCntEnvSus_0;
				_egCountStep_0 = _egCountStepDR_0;
				return _egState_0 = 2;
			}
			// -> SUSTAIN
			_egGate_0 = 0x7fffffff;
			_egCountStep_0 = 0;
			return _egState_0 = 4;
		// DECAY ->
		case 2:
			// -> SUSTAIN
			if ((_egTL_0 + _egSL_0) < _EG_SUSTAIN_BTM)
			{
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				_egCount_0 = _egCntEnvSus_0;
				return _egState_0 = 4;
			}
		// DECAY,RELEASE,SoundOFF -> EG OFF
		default: return _egState_0 = 0;
	}
}
int32 SynthesizerFM2::VoiceUnitFM2::_shiftEGState_1()
{
	switch (_egState_1)
	{
		// ATTACK ->
		case 1:
			_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
			_egCount_1 = 0;
			// -> DECAY
			if (_egSL_1 > 0)
			{
				_egGate_1 =  _egCntEnvSus_1;
				_egCountStep_1 = _egCountStepDR_1;
				return _egState_1 = 2;
			}
			// -> SUSTAIN
			_egGate_1 = 0x7fffffff;
			_egCountStep_1 = 0;
			return _egState_1 = 4;
		// DECAY ->
		case 2:
			// -> SUSTAIN
			if ((_egTL_1 + _egSL_1) < _EG_SUSTAIN_BTM)
			{
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				_egGate_1 = 0x7fffffff;
				_egCountStep_1 = 0;
				_egCount_1 = _egCntEnvSus_1;
				return _egState_1 = 4;
			}
		// DECAY,RELEASE,SoundOFF -> EG OFF
		default: return _egState_1 = 0;
	}
}
void SynthesizerFM2::VoiceUnitFM2::keyOn(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0x0013;
	partNumber = ch;
	noteNumber = note;
// tone ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	TonePresetFM* _preset = nullptr;
	int _keyFlw_0 = egPow2tIdx, 
	    _keyFlw_1 = egPow2tIdx;
	if (_part.mode == 2)
	{
		_preset = &TONE_PRESETS[noteNumber | 128];
		_pgKey_0 = _POW2T_KEY_TABLE[_preset->key] + _MUL_TABLE[_preset->mul_0] + (_preset->tune_0 - 8192) + pgKeyPow2tIdx;
		_pgKey_1 = _POW2T_KEY_TABLE[_preset->key] + _MUL_TABLE[_preset->mul_1] + (_preset->tune_1 - 8192) + pgKeyPow2tIdx;
		//_keyFlw = 0;
		_prPan = _preset->pan;
	}
	else
	{
		_preset = &TONE_PRESETS[_part.programChange];
		_pgKey_0 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset->mul_0] + (_preset->tune_0 - 8192) + pgKeyPow2tIdx;
		_pgKey_1 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset->mul_1] + (_preset->tune_1 - 8192) + pgKeyPow2tIdx;
		_keyFlw_0 += _EG_KEY_TABLE[127 + noteNumber - _preset->kfStart_0] << _preset->kf_0;
		_keyFlw_1 += _EG_KEY_TABLE[127 + noteNumber - _preset->kfStart_1] << _preset->kf_1;
		_prPan = 64;
	}
// LFO ---
	_lfoWaveTable = &_LFO_WAVE_TABLES[_preset->lfoWave << _LFO_WAVE_TABLE_SIZE_BITS];
	_lfoModSens_0 = _preset->lms_0;
	_lfoModSens_1 = _preset->lms_1;
	_lfoMod_0 = _part.ccValues[1] * _lfoModSens_0;
	_lfoMod_1 = _part.ccValues[1] * _lfoModSens_1;
	_lfoTimerStep = _preset->lfoFreq * lfoTimerCoef;
	_lfoTimer = _LFO_TIMER_INIT_VALUE;
	_lfoPhase = _preset->strc & 1;
	_lfoOutPM_0 = 0;
	_lfoOutPM_1 = 0;
	_lfoOutAM_0 = 0;
	_lfoOutAM_1 = 0;
// Velocity Control ---
	_egTL_0 = _preset->tl_0 + ((127 - velo) >> (7 - _preset->tlVelCtrl_0));
	if (_egTL_0 > 127)_egTL_0 = 0;
	_egTL_1 = _preset->tl_1 + ((127 - velo) >> (7 - _preset->tlVelCtrl_1));
	if (_egTL_1 > 127)_egTL_1 = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// pitch ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
// process ---
	process = _PROC_PTRS[_preset->strc];
	_fdbk = _EG_TOTAL_LEVEL_TABLE[_preset->fl];
// Operator 0 --->
	_waveTable_0 = &WAVE_TABLES[_preset->wave_0 << _WAVE_TABLE_SIZE_BITS];
// PG ---
	int v = _pgKey_0 + _pgBend;
	_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase_0 = 0;
// EG ---
	_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
	_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
	_egSL_0 = _preset->sl_0;
	_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
	if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
	v = (_preset->dr_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
	_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset->rr_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
	_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset->ar_0 < 127) {
		_egCount_0 = _EG_CNT_ENV_ATK;
		_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
		_egGate_0 = _EG_MAX_CNT;
		v = (_preset->ar_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
		_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState_0 = 1;
	} else {
		_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
		_egCount_0 = 0;
	// -> DECAY
		if (_egSL_0 > 0) {
			_egGate_0 = _egCntEnvSus_0;
			_egCountStep_0 = _egCountStepDR_0;
			_egState_0 = 2;
	// -> SUSTAIN
		} else {
			_egGate_0 = 0x7fffffff;
			_egCountStep_0 = 0;
			_egState_0 = 4; }
	}
// Operator 1 --->
	_waveTable_1 = &WAVE_TABLES[_preset->wave_1 << _WAVE_TABLE_SIZE_BITS];
// PG ---
	v = _pgKey_1 + _pgBend;
	_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase_1 = 0;
// EG ---
	_egTLv_1 = _EG_TOTAL_LEVEL_TABLE[_egTL_1];
	_egCntEnvBtm_1 = _EG_CNT_GATE_TABLE[128 - _egTL_1];
	_egSL_1 = _preset->sl_1;
	_egCntEnvSus_1 = _EG_CNT_GATE_TABLE[_egSL_1];
	if (_egCntEnvSus_1 > _egCntEnvBtm_1)_egCntEnvSus_1 = _egCntEnvBtm_1;
	v = (_preset->dr_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
	_egCountStepDR_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset->rr_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
	_egCountStepRR_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset->ar_1 < 127) {
		_egCount_1 = _EG_CNT_ENV_ATK;
		_egLevelTable_1 = _EG_LEVEL_TABLE_UP;
		_egGate_1 = _EG_MAX_CNT;
		v = (_preset->ar_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
		_egCountStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState_1 = 1;
	} else {
		_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
		_egCount_1 = 0;
	// -> DECAY
		if (_egSL_1 > 0) {
			_egGate_1 = _egCntEnvSus_1;
			_egCountStep_1 = _egCountStepDR_1;
			_egState_1 = 2;
	// -> SUSTAIN
		} else {
			_egGate_1 = 0x7fffffff;
			_egCountStep_1 = 0;
			_egState_1 = 4; }
	}
}
void SynthesizerFM2::VoiceUnitFM2::keyOnSampler(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0x0011;
	partNumber = ch;
	noteNumber = note;
// tone ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	SamplerPreset& _preset = SAMPLER_PRESETS[noteNumber];
	SampleData& _sample = SAMPLE_DATAS[_preset.smplNum];
	if (_sample.sampleData == nullptr){ killProcess(); return; }
	_waveTable_0 = _sample.sampleData;
	_plbkE = _sample.plbkE;
	_loopE = _sample.loopE;
	_loopF = _sample.loopF;
	_loopFk = _sample.loopFk;
// LFO ---
	_lfoOutPM_0 = 0;
	_lfoOutAM_0 = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_prPan = _preset.pan;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// PG ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	_pgKey_0 = ((_preset.octv - 8) << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
	int v = _pgKey_0 + _pgBend;
	_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase_0 = _sample.plbkS;
	_pgPhaseAFP = 0;
	_pgKey_1 = 0; _pgPhase_1 = 0;
// EG ---
	_egTL_0 = _preset.tl;
	_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
	_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
	_egSL_0 = _preset.sl;
	_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
	if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
	v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset.ar < 127) {
		_egCount_0 = _EG_CNT_ENV_ATK;
		_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
		_egGate_0 = _EG_MAX_CNT;
		v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState_0 = 1;
	} else {
		_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
		_egCount_0 = 0;
	// -> DECAY
		if (_egSL_0 > 0) {
			_egGate_0 = _egCntEnvSus_0;
			_egCountStep_0 = _egCountStepDR_0;
			_egState_0 = 2;
	// -> SUSTAIN
		} else {
			_egGate_0 = 0x7fffffff;
			_egCountStep_0 = 0;
			_egState_0 = 4; }
	}
// process ---
	process = (_sample.channelCount == 2)? _PROC_PTRS[9] : _PROC_PTRS[8];
}
void SynthesizerFM2::VoiceUnitFM2::updPartBendSens(const int32& val)
{
		_ptBendSens = _POW2T_KEY_TABLE[val];
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey_0 + _pgBend + _lfoOutPM_0;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		// Inst op2 --->
		if (state & 2)
		{
			v = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
}
void SynthesizerFM2::VoiceUnitFM2::updPartPitchBend(const int32& val)
{
		_ptBend = val - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey_0 + _pgBend + _lfoOutPM_0;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		// Inst op2 --->
		if (state & 2)
		{
			v = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
}
void SynthesizerFM2::VoiceUnitFM2::updPartModulation(const int32& val)
{
		_lfoMod_0 = val * _lfoModSens_0;
		_lfoMod_1 = val * _lfoModSens_1;
}
void SynthesizerFM2::VoiceUnitFM2::updateLfoFrequency(const int32& val)
{
	_lfoTimerStep = val * lfoTimerCoef;
}
void SynthesizerFM2::VoiceUnitFM2::_proc0(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
		// pg update --->
			x = _wt_0[_pgPhase_0 >> _PG_PHASE_FP] + x;
		// OP 0 out --->
			n = _log2vT[x];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else
				{
					if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
		// OP 1 out ----->
			n += _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc1(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
		// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
		// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
		// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			n = _log2vT[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else
				{
					if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
		// OP 1 out ----->
			n += _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc2(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			// op0
			_lfoOutPM_0 = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
		// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
		// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
		// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			n = _log2vT[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else
				{
					if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
		// OP 1 out ----->
			n += _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc3(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			x = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
		// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
		// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
		// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			n = _log2vT[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else
				{
					if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
		// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
		// OP 1 out ----->
			n += _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc4(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
		// pg update --->
			x = _wt_0[_pgPhase_0 >> _PG_PHASE_FP] + x;
		// OP 0 out --->
			p = _log2vTI[x];
		// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else p = 0;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else { killProcess(); return; }
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			n = _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc5(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
		// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
		// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
		// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			p = _log2vTI[p];
		// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else p = 0;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else { killProcess(); return; }
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			n = _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc6(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			// op0
			_lfoOutPM_0 = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
		// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
		// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
		// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			p = _log2vTI[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else p = 0;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else { killProcess(); return; }
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			n = _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc7(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, *const&_wt_1 = _waveTable_1;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			x = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 1)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et_0 = _egLevelTable_0;
				else
				{
					if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
		// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
		// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
		// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			p = _log2vTI[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else p = 0;
	// Operator 1 --->
		if (state & 2)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState_1())_et_1 = _egLevelTable_1;
				else { killProcess(); return; }
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
		// pg update --->
			p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			n = _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc8(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable_0;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable_0;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount_0 >= _egGate_0)
		{
			if (_shiftEGState_0())_et = _egLevelTable_0;
			else { killProcess(); return; }
		}
		int v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
	// pg update --->
		if (_pgPhase_0 >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase_0] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount_0 += _egCountStep_0;
		_pgPhaseAFP += _pgPhaseStep_0;
		_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerFM2::VoiceUnitFM2::_proc9(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable_0;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable_0;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount_0 >= _egGate_0)
		{
			if (_shiftEGState_0())_et = _egLevelTable_0;
			else { killProcess(); return; }
		}
		int v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
	// pg update --->
		if (_pgPhase_0 >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase_0 << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_egCount_0 += _egCountStep_0;
		_pgPhaseAFP += _pgPhaseStep_0;
		_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerFM2::VoiceUnitFM2::setup(SynthData* synthData)
{
	// volume Table
	_mainVolTable = synthData->getVOLUME_TABLE_0dB();
	_exprVolTable = synthData->getVOLUME_TABLE_0dB();
	_veloVolTable = synthData->getVOLUME_TABLE_0dB();
	_panTable = synthData->getPAN_TABLE_SIN();
	// Log2 Volume Table
	_LOG2V_TABLE = synthData->getLOG2V_TABLE();
	_LOG2V_TABLE_I = synthData->getLOG2V_TABLE_I();
	// Pow2 Table
	_POW2_TABLE = synthData->getPOW2_TABLE();
	_POW2T_RS_TABLE = synthData->getPOW2T_RS_TABLE();
	_POW2T_KEY_TABLE = synthData->getPOW2T_KEY_TABLE();
	// EG
	_EG_TOTAL_LEVEL_TABLE = synthData->getEG_TOTAL_LEVEL_TABLE();
	_EG_LEVEL_TABLE_UP = synthData->getEG_LEVEL_TABLE_UP();
	_EG_LEVEL_TABLE_DW = synthData->getEG_LEVEL_TABLE_DW();
	_EG_KEY_TABLE = synthData->getEG_KEY_TABLE();
	_EG_POW2T_RS_TABLE = synthData->getEG_POW2T_RS_TABLE();
	_EG_CNT_GATE_TABLE = synthData->getEG_CNT_GATE_TABLE();
	// etc
	_MUL_TABLE = synthData->getMUL_TABLE();
	_LFO_WAVE_TABLES = synthData->getLFO_WAVE_TABLES();
}

}}}
